########################################################################
#
#   Tryton account_invoice_do module for Dominican Republic
#   Copyright (C) 2014 Franky Almonte <falmonte@fdsl.org.do>
#   Copyright (C) 2014 Fundacion Dominicana de Software Libre
#                       <tryton@fdsl.org.do>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#########################################################################
from decimal import Decimal
import base64
import itertools
import operator
from sql import Literal
from sql.aggregate import Count, Sum
from sql.conditionals import Coalesce, Case
from sql.functions import Abs, Sign

from trytond.model import Workflow, ModelView, ModelSQL, fields
from trytond.report import Report
from trytond.wizard import Wizard, StateView, StateTransition, StateAction, \
    Button
from trytond import backend
from trytond.pyson import If, Eval, Bool, Id
from trytond.tools import reduce_ids
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.rpc import RPC

__all__ = ['InvoiceNCF']

#_STATES = {
#    'readonly': Eval('state') != 'draft',
#}
#_DEPENDS = ['state']

#_TYPE = [
#    ('out_invoice', 'Invoice'),
#    ('in_invoice', 'Supplier Invoice'),
#    ('out_credit_note', 'Credit Note'),
#    ('in_credit_note', 'Supplier Credit Note'),
#]

_NCF_TYPE = [
        ('out_ncf_credito_fiscal', 'Factura que Genera Credito Fiscal'),
        ('out_ncf_consumidor_final', 'Factura para Consumidor Final'),
        ('out_ncf_nota_debito', 'Nota de Debito'),
        ('out_ncf_nota_credito', 'Nota de Credito'),
        ('out_ncf_especial_proveedor_informal', 'Factura para Proveedores Informales'),
        ('out_ncf_especial_ingreso_unico', 'Factura para Registro Unico de Ingresos'),
        ('out_ncf_especial_regimen_especial', 'Factura para Regimenes Especiales de Tributacion'),
        ('out_ncf_especial_gubernamental', 'Factura para Institucion Gubernamental'),
]

#_NCFTYPE2JOURNAL = {
#    'out_ncf_credito_fiscal': 'revenue',
#    'out_ncf_consumidor_final': 'revenue',
#    'out_ncf_nota_debito': 'revenue',
#    'out_ncf_nota_credito': 'revenue',
#    'out_ncf_especial_proveedor_informal': 'revenue',
#    'out_ncf_especial_ingreso_unico': 'revenue',
#    'out_ncf_especial_regimenes_especiales': 'revenue',
#    'out_ncf_especial_gubernamental': 'revenue',
#}

#_ZERO = Decimal('0.0')

#_CREDIT_TYPE = {
#    None: None,
#    'out_invoice': 'out_credit_note',
#    'in_invoice': 'in_credit_note',
#    'out_credit_note': 'out_invoice',
#    'in_credit_note': 'in_invoice',
#    }


class InvoiceNCF(Workflow, ModelSQL, ModelView):
    'Invoice'
    __name__ = 'account.invoice'
    ncf_type = fields.Selection(_NCF_TYPE, 'NCF Type', select=True,
        required=True, states={
            'readonly': ((Eval('state') != 'draft')
                | Eval('context', {}).get('ncf_type'))
            }, depends=['state'], help='Choose a NCF type')
    ncf_name = fields.Function(fields.Char('NCF Type'), 'get_ncf_type_name')
    ncf_number = fields.Char('NCF Number', size=19, readonly=True, select=True)

    @classmethod
    def __setup__(cls):
        super(InvoiceNCF, cls).__setup__()
        cls._error_messages.update({
            'no_invoice_ncf_sequence': ('There is no invoice NCF sequence for '
                'invoice "%(invoice)s" on the period/fiscal year '
                '"%(period)s".')
            })

    @staticmethod
    def default_ncf_type():
        return Transaction().context.get('ncf_type', 'out_ncf_consumidor_final')

#    @fields.depends('ncf_type', 'party', 'company')
#    def on_change_ncf_type(self):
#        Journal = Pool().get('account.journal')
#        res = {}
#        journals = Journal.search([
#                ('type', '=', _NCFTYPE2JOURNAL.get(self.ncf_type or 'out_invoice_ncf_consumidor_final',
#                        'revenue')),
#                ], limit=1)
#        if journals:
#            journal, = journals
#            res['journal'] = journal.id
#            res['journal.rec_name'] = journal.rec_name
#        res.update(Invoice.__get_account_payment_term())
#        return res

    @classmethod
    def get_ncf_type_name(self, invoices, name):
        ncf_type_names = {}
        ncf_type2name = {}
        for ncf_type, ncf_name in self.fields_get(fields_names=['ncf_type']
                )['ncf_type']['selection']:
            ncf_type2name[ncf_type] = ncf_name
        for invoice in invoices:
            ncf_type_names[invoice.id] = ncf_type2name[invoice.ncf_type]
        return ncf_type_names


    def set_ncf_number(self):
        '''
        Set NCF number to the invoice
        '''
        pool = Pool()
        Period = pool.get('account.period')
        Sequence = pool.get('ir.sequence.strict')
        Date = pool.get('ir.date')

        if self.ncf_number:
            return

        test_state = True
        if self.type in ('in_invoice', 'in_credit_note'):
            test_state = False

        accounting_date = self.accounting_date or self.invoice_date
        period_id = Period.find(self.company.id,
            date=accounting_date, test_state=test_state)
        period = Period(period_id)
        sequence = period.get_ncf_sequence(self.ncf_type)
        if not sequence:
            self.raise_user_error('no_invoice_ncf_sequence', {
                    'invoice': self.rec_name,
                    'period': period.rec_name,
                    })
        with Transaction().set_context(
                date=self.invoice_date or Date.today()):
            ncf_number = Sequence.get_id(sequence.id)
            vals = {'ncf_number': ncf_number}
            if (not self.invoice_date
                    and self.ncf_type in ('out_ncf_credito_fiscal', 'out_ncf_consumidor_final',
                        'out_ncf_nota_credito', 'out_ncf_especial_proveedor_informal',
                        'out_ncf_especial_ingreso_unico', 'out_ncf_especial_regimen_especial',
                        'out_ncf_gubernamental')):
                vals['invoice_date'] = Transaction().context['date']
        self.write([self], vals)


    @classmethod
    @ModelView.button
    @Workflow.transition('posted')
    def post(cls, invoices):
        Move = Pool().get('account.move')

        moves = []
        for invoice in invoices:
            invoice.set_number()
            invoice.set_ncf_number()
            moves.append(invoice.create_move())
        cls.write(invoices, {
                'state': 'posted',
                })
        Move.post(moves)
        for invoice in invoices:
            if invoice.type in ('out_invoice', 'out_credit_note'):
                invoice.print_invoice()
